from typing import List, Optional, Union

import pandas as pd

from intake.source import base

from . import __version__
from .utils import (
    load_metadata,
    make_data_url,
    make_filter,
    make_label,
    make_metadata_url,
    make_search_docs_url,
    response_from_url,
)


class AXDSSensorSource(base.DataSource):
    """Intake Source for AXDS sensor

    Parameters
    ----------
    internal_id : Optional[int], optional
        Internal station id for Axiom, by default None. Not the UUID or dataset_id. Need to input internal_id or dataset_id. If both are input, be sure they are for the same station.
    dataset_id : Optional[str], optional
        The UUID for the station, by default None. Not the internal_id. Need to input internal_id or dataset_id. If both are input, be sure they are for the same station.
    start_time : Optional[str], optional
        At what datetime for data to start, by default None. Must be interpretable by pandas ``Timestamp``. If not input, the datetime at which the dataset starts will be used.
    end_time : Optional[str], optional
        At what datetime for data to end, by default None. Must be interpretable by pandas ``Timestamp``. If not input, the datetime at which the dataset ends will be used.
    qartod : bool, int, list, optional
        Whether to return QARTOD agg flags when available, which is only for sensor_stations. Can instead input an int or a list of ints representing the _qa_agg flags for which to return data values. More information about QARTOD testing and flags can be found here: https://cdn.ioos.noaa.gov/media/2020/07/QARTOD-Data-Flags-Manual_version1.2final.pdf. Only used by datatype "sensor_station". Is not available if `binned==True`.

        Examples of ways to use this input are:

        * ``qartod=True``: Return aggregate QARTOD flags as a column for each data variable.
        * ``qartod=False``: Do not return any QARTOD flag columns.
        * ``qartod=1``: nan any data values for which the aggregated QARTOD flags are not equal to 1.
        * ``qartod=[1,3]``: nan any data values for which the aggregated QARTOD flags are not equal to 1 or 3.

        Flags are:

        * 1: Pass
        * 2: Not Evaluated
        * 3: Suspect
        * 4: Fail
        * 9: Missing Data

    use_units : bool, optional
        If True include units in column names. Syntax is "standard_name [units]". If False, no units. Then syntax for column names is "standard_name". This is currently specific to sensor_station only. Only used by datatype "sensor_station".
    metadata : dict, optional
        Metadata for catalog.
    binned : bool, optional
        True for binned data, False for raw, by default False. Only used by datatype "sensor_station".
    bin_interval : Optional[str], optional
        If ``binned=True``, input the binning interval to return. Options are hourly, daily, weekly, monthly, yearly. If bin_interval is input, binned is set to True. Only used by datatype "sensor_station".
    only_pgids : list, optional
        If input, only return data associated with these parameterGroupIds. This is separate from parameterGroupLabels and parameterGroupIds that might be present in the metadata.

    Raises
    ------
    ValueError
        _description_
    """

    name = "axds-sensor"
    version = __version__
    container = "dataframe"
    partition_access = True

    def __init__(
        self,
        internal_id: Optional[int] = None,
        dataset_id: Optional[str] = None,
        start_time: Optional[str] = None,
        end_time: Optional[str] = None,
        qartod: Union[int, List[int], bool] = False,
        use_units: bool = True,
        metadata=None,
        binned: bool = False,
        bin_interval: Optional[str] = None,
        only_pgids: Optional[List[int]] = None,
    ):

        if internal_id is None and dataset_id is None:
            raise ValueError(
                "internal_id and dataset_id cannot both be None. Input one of them."
            )

        self.dataset_id = dataset_id
        self.start_time = start_time
        self.end_time = end_time
        self.internal_id = internal_id
        self.qartod = qartod
        self.use_units = use_units
        self.only_pgids = only_pgids

        if bin_interval is not None:
            binned = True
        self.binned = binned
        self.bin_interval = bin_interval

        if self.binned:
            if self.qartod:
                raise ValueError(
                    "QARTOD is not available for binned output. Set QARTOD to False or use raw data."
                )

        # need dataset_id to get metadata
        if self.dataset_id is None:
            assert self.internal_id is not None
            res = response_from_url(make_metadata_url(make_filter(self.internal_id)))
            assert isinstance(res, dict)
            self.dataset_id = res["data"]["stations"][0]["uuid"]

        # need internal_id to get data
        elif self.internal_id is None:
            assert self.dataset_id is not None
            self.internal_id = response_from_url(make_search_docs_url(self.dataset_id))[
                0
            ]["id"]

        self._dataframe = None

        metadata = metadata or {}
        metadata["dataset_id"] = self.dataset_id

        # this is what shows in the source if you print it
        self._captured_init_kwargs.update(
            {
                "internal_id": self.internal_id,
                "dataset_id": self.dataset_id,
                "start_time": self.start_time,
                "end_time": self.end_time,
                "qartod": self.qartod,
                "use_units": self.use_units,
                "binned": self.binned,
                "bin_interval": bin_interval,
                "only_pgids": self.only_pgids,
            }
        )

        super(AXDSSensorSource, self).__init__(metadata=metadata)

    def get_filters(self):
        """Return appropriate filter for stationid.

        What filter form to use depends on if V1 or V2.

        For V1, use each parameterGroupId only once to make a filter since all data of that type will be read in together.

        Following Sensor API
        https://admin.axds.co/#!/sensors/api/overview
        """

        filters = []

        if self.metadata["version"] == 1:

            # if we should only return the requested variables, use the pgids input for this.
            if self.only_pgids:
                pgids = self.only_pgids
            else:
                pgids = [
                    self.metadata["variables_details"][var]["parameterGroupId"]
                    for var in self.metadata["variables_details"]
                ]
            for pgid in list(set(pgids)):
                filters.append(make_filter(self.internal_id, pgid))
        else:
            filters.append(make_filter(self.internal_id))
        return filters

    def _load_to_dataframe(self, url: str) -> pd.DataFrame:
        """load from raw data url to DataFrame.

        For V1 stations the result of a call to this function will be one of potentially many calls for data, but there will be only one loop below in a call.
        For V2 stations the result of a call to this function will be the only call for data, but there may be several loops in the call.

        Parameters
        ----------
        url : str
            URL to find data from.

        Returns:
        DataFrame
            Data.
        """
        data_raw = response_from_url(url)

        # check for presence of any data
        assert isinstance(data_raw, dict)
        # for binned data
        if "data" not in data_raw:
            self._dataframe = None
            raise ValueError(f"No data found for url {url}.")

        # for raw data
        if len(data_raw["data"]["groupedFeeds"]) == 0:
            self._dataframe = None
            raise ValueError(f"No data found for url {url}.")

        # loop over the data feeds and read the data into DataFrames
        # link to other metadata as needed
        dfs = []
        for feed in data_raw["data"]["groupedFeeds"]:

            # different names for data sets depending on if binned or not
            if self.binned:
                metadata_values_name = "avgVals"
            else:
                metadata_values_name = "values"

            # for the case where we only return data from certain pgids, skip loop if this
            # data is not included
            if self.only_pgids is not None:
                if not any(
                    [
                        var["parameterGroupId"]
                        for var in feed["metadata"][metadata_values_name]
                        if var["parameterGroupId"] in self.only_pgids
                    ]
                ):
                    continue

            # import pdb; pdb.set_trace()

            columns = {}  # all non-index columns in dataframe
            indices = {}  # indices for dataframe

            # indices first: time and potentially z
            feed_ind = feed["metadata"]["time"]
            feed_ind["column_name"] = make_label(
                feed_ind["label"], feed_ind.get("units", None), use_units=self.use_units
            )
            indices[feed_ind["index"]] = feed_ind

            # in this case, z or depth is also an index
            if feed["metadata"]["z"] is not None:
                feed_ind = feed["metadata"]["z"]
                feed_ind["column_name"] = make_label(
                    feed_ind["label"],
                    feed_ind.get("units", None),
                    use_units=self.use_units,
                )
                indices[feed_ind["index"]] = feed_ind

            # These should never be non-None for sensors
            if (
                feed["metadata"]["lon"] is not None
                or feed["metadata"]["lat"] is not None
            ):
                lon, lat = feed["metadata"]["lon"], feed["metadata"]["lat"]
                raise ValueError(
                    f"lon/lat should be None for sensors but are {lon}, {lat}."
                )

            # add data columns
            data_cols = {
                val["index"]: val for val in feed["metadata"][metadata_values_name]
            }
            # match variable name from metadata (standard_name) to be column name
            for index in data_cols:
                # variable name
                label = [
                    var
                    for var in self.metadata["variables_details"]
                    if self.metadata["variables_details"][var]["deviceId"]
                    == data_cols[index]["deviceId"]
                ][0]
                data_cols[index]["label"] = label
                # column name
                data_cols[index]["column_name"] = make_label(
                    label, data_cols[index].get("units", None), use_units=self.use_units
                )

            columns.update(data_cols)

            # whether or not to read in QARTOD Aggregation flags is chosen at the catalog level in axds_cat.py
            # columns only includes the QA column info if qartod is True
            # or, include QARTOD columns but then remove data rows based on the flag values.
            qartod = self.qartod
            if isinstance(qartod, (str, list)) or qartod:

                # add qartod columns
                qa_cols = {val["index"]: val for val in feed["metadata"]["qcAgg"]}
                # match variable name using deviceId from metadata (standard_name) to be column name
                for index in qa_cols:
                    label = [
                        var
                        for var in self.metadata["variables_details"]
                        if self.metadata["variables_details"][var]["deviceId"]
                        == qa_cols[index]["deviceId"]
                    ][0]
                    qa_cols[index]["label"] = f"{label}_qc_agg"
                    qa_cols[index]["column_name"] = qa_cols[index]["label"]

                    # match deviceId between data_col and qa_col to get column name associated with qa_col
                    name = [
                        data_cols[ind]["column_name"]
                        for ind in data_cols
                        if data_cols[ind]["deviceId"] == qa_cols[index]["deviceId"]
                    ][0]
                    qa_cols[index]["data_name"] = name
                    columns.update(qa_cols)

            col_names = [columns[i]["column_name"] for i in list(columns)]
            ind_names = [indices[i]["column_name"] for i in list(indices)]

            # do this in steps in case we are dropping QA columns
            df = pd.DataFrame(feed["data"])
            icolumns_to_keep = list(indices) + list(columns)
            df = df[icolumns_to_keep]
            df.columns = ind_names + col_names
            df.set_index(ind_names, inplace=True)

            # nan data for which QARTOD flags shouldn't be included
            if isinstance(qartod, (str, list)):
                if isinstance(qartod, str):
                    qartod = [qartod]

                for ind in qa_cols:
                    data_name = qa_cols[ind]["data_name"]  # data column name
                    qa_name = qa_cols[ind]["column_name"]  # qa column name

                    for qa in qartod:
                        df.loc[df[qa_name] != qa, data_name] = pd.NA

                # drop qartod columns
                df.drop(labels=qa_name, axis=1, inplace=True)

            dfs.append(df)

        df = dfs[0]

        # this gets different and I think better results than dfs[0].join(dfs[1:], how="outer", sort=True)
        # even though they should probably return the same thing.
        for i in range(1, len(dfs)):
            df = df.join(dfs[i], how="outer", sort=True)

        return df

    @property
    def data_urls(self):
        """Prepare to load in data by getting data_urls.

        For V1 sources there will be a data_url per parameterGroupId but not for V2 sources.
        """

        if not hasattr(self, "_data_urls"):

            # get extended metadata which we need both for reading the data and as metadata
            result = response_from_url(make_search_docs_url(self.dataset_id))[0]
            self.metadata.update(load_metadata("sensor_station", result))

            start_time = self.start_time or self.metadata["minTime"]
            end_time = self.end_time or self.metadata["maxTime"]

            filters = self.get_filters()
            self._data_urls = [
                make_data_url(
                    filter, start_time, end_time, self.binned, self.bin_interval
                )
                for filter in filters
            ]
        return self._data_urls

    def _load(self):
        """How to load in a specific station once you know it by dataset_id"""

        dfs = [self._load_to_dataframe(url) for url in self.data_urls]

        df = dfs[0]
        # this gets different and I think better results than dfs[0].join(dfs[1:], how="outer", sort=True)
        # even though they should probably return the same thing.
        for i in range(1, len(dfs)):
            df = df.join(dfs[i], how="outer", sort=True)

        self._dataframe = df

    def _get_schema(self):
        """get schema"""
        if self._dataframe is None:
            # TODO: could do partial read with chunksize to get likely schema from
            # first few records, rather than loading the whole thing
            self._load()
        return base.Schema(
            datashape=None,
            dtype=self._dataframe.dtypes,
            shape=self._dataframe.shape,
            npartitions=1,
            extra_metadata={},
        )

    def _get_partition(self, _):
        """get partition"""
        if self._dataframe is None:
            self._load_metadata()
        return self._dataframe

    def read(self):
        """read data in"""
        return self._get_partition(None)

    def _close(self):
        """close source"""
        self._dataframe = None



"""Utils to run."""

from importlib.metadata import PackageNotFoundError, version
from operator import itemgetter
from typing import Optional, Union

import cf_pandas as cfp
import pandas as pd
import requests

from shapely import wkt


search_headers = {"Accept": "application/json"}
baseurl = "https://sensors.axds.co/api"


def _get_version() -> str:
    """Fixes circular import issues."""
    try:
        __version__ = version("ocean-model-skill-assessor")
    except PackageNotFoundError:
        # package is not installed
        __version__ = "unknown"

    return __version__


def return_parameter_options() -> dict:
    """Find parameters for Axiom assets.

    Returns
    -------
    List
        Contains the parameter information for Axiom assets.

    Examples
    --------
    >>> return_parameter_options()
    [{'id': 4,
    'label': 'Relative Humidity',
    'urn': 'http://mmisw.org/ont/cf/parameter/relative_humidity',
    'ratio': False,
    'sanityMin': 0.0,
    'sanityMax': 110.0,
    'parameterGroupDefault': True,
    'configJson': None,
    'stageConfigJson': None,
    'idSanityUnit': 1,
    'idParameterGroup': 22,
    'idParameterType': 101,
    'parameterName': 'relative_humidity'},
    ...
    """

    resp = requests.get("http://oikos.axds.co/rest/context")
    # resp.raise_for_status()
    output = resp.json()
    # params = data["parameters"]

    return output


def available_names() -> list:
    """Return available parameterNames for variables.

    Returns
    -------
    list
        parametersNames, which are a superset of standard_names.
    """

    resp = return_parameter_options()
    params = resp["parameters"]

    # find parameterName options for AXDS. These are a superset of standard_names
    names = [i["parameterName"] for i in params]

    return names


def match_key_to_parameter(
    keys_to_match: list,
    criteria: Optional[dict] = None,
) -> list:
    """Find Parameter Group values that match keys_to_match.

    Parameters
    ----------
    keys_to_match : list
        The custom_criteria key to narrow the search, which will be matched to the category results
        using the custom_criteria that must be set up ahead of time with `cf-pandas`.
    criteria : dict, optional
        Criteria to use to map from variable to attributes describing the variable. If user has
        defined custom_criteria, this will be used by default.

    Returns
    -------
    list
        Parameter Group values that match key, according to the custom criteria.
    """

    resp = return_parameter_options()
    params = resp["parameters"]

    # find parameterName options for AXDS. These are a superset of standard_names
    names = [i["parameterName"] for i in params]
    group_params = resp["parameterGroups"]

    # select parameterName that matches selected key
    vars = cfp.match_criteria_key(names, keys_to_match, criteria)

    # find parametergroupid that matches var
    pgids = [
        i["idParameterGroup"]
        for var in vars
        for i in params
        if i["parameterName"] == var
    ]

    # find parametergroup label to match id
    pglabels = [i["label"] for pgid in pgids for i in group_params if i["id"] == pgid]

    # want unique but ordered returned
    return list(zip(*sorted(set(zip(pglabels, pgids)))))

    # return pglabels, pgids
    # return list(set(pglabels))


def match_std_names_to_parameter(standard_names: list) -> list:
    """Find Parameter Group values that match standard_names.

    Parameters
    ----------
    standard_names : list
        standard_names values to narrow the search.

    Returns
    -------
    list
        Parameter Group values that match standard_names.
    """

    resp = return_parameter_options()
    params = resp["parameters"]

    names = [i["parameterName"] for i in params]

    if not all([std_name in names for std_name in standard_names]):
        raise ValueError(
            """Input standard_names are not all matches with system parameterNames.
                          Check available values with `intake_axds.available_names()`."""
        )

    group_params = resp["parameterGroups"]

    # find parametergroupid that matches std_name
    pgids = [
        i["idParameterGroup"]
        for std_name in standard_names
        for i in params
        if i["parameterName"] == std_name
    ]

    # find parametergroup label to match id
    pglabels = [i["label"] for pgid in pgids for i in group_params if i["id"] == pgid]

    # want unique but ordered returned
    return list(zip(*sorted(set(zip(pglabels, pgids)))))

    # return pglabels, pgids
    # return list(set(pglabels))


def load_metadata(datatype: str, results: dict) -> dict:  #: Dict[str, str]
    """Load metadata for catalog entry.

    Parameters
    ----------
    results : dict
        Returned results from call to server for a single dataset.

    Returns
    -------
    dict
        Metadata to store with catalog entry.
    """

    # matching names in intake-erddap
    keys = ["datasetID", "title", "summary", "type", "minTime", "maxTime"]
    # names of keys in Axiom system.
    items = [
        "uuid",
        "label",
        "description",
        "type",
        "start_date_time",
        "end_date_time",
    ]
    values = itemgetter(*items)(results)
    metadata = dict(zip(keys, values))

    if datatype == "platform2":
        metadata["institution"] = (
            results["source"]["meta"]["attributes"]["institution"]
            if "institution" in results["source"]["meta"]["attributes"]
            else None
        )
        metadata["geospatial_bounds"] = results["source"]["meta"]["attributes"][
            "geospatial_bounds"
        ]

        p1 = wkt.loads(metadata["geospatial_bounds"])
        keys = ["minLongitude", "minLatitude", "maxLongitude", "maxLatitude"]
        metadata.update(dict(zip(keys, p1.bounds)))

        # save variable details if they have a standard_name
        # some platforms have lots of variables that seem irrelevant
        out = {
            attrs["attributes"]["standard_name"]: {
                "variable_name": varname,
                "units": attrs["attributes"]["units"]
                if "units" in attrs["attributes"]
                else None,
                "unit_id": attrs["attributes"]["unit_id"]
                if "unit_id" in attrs["attributes"]
                else None,
                "long_name": attrs["attributes"]["long_name"],
                "parameter_id": attrs["attributes"]["parameter_id"]
                if "parameter_id" in attrs["attributes"]
                else None,
            }
            for varname, attrs in results["source"]["meta"]["variables"].items()
            if "standard_name" in attrs["attributes"]
        }

        metadata["variables_details"] = out
        metadata["variables"] = list(out.keys())

    elif datatype == "sensor_station":

        # INSTITUTION?
        # location is lon, lat, depth and type
        # e.g. {'coordinates': [-123.711083, 38.914556, 0.0], 'type': 'Point'}
        lon, lat, depth = results["data"]["location"]["coordinates"]
        keys = ["minLongitude", "minLatitude", "maxLongitude", "maxLatitude"]
        metadata.update(dict(zip(keys, [lon, lat, lon, lat])))

        # e.g. 106793
        metadata["internal_id"] = results["data"]["id"]

        # variables, standard_names (or at least parameterNames)
        figs = results["data"]["figures"]

        out = {
            subPlot["datasetVariableId"]: {
                "parameterGroupLabel": fig["label"],
                "parameterGroupId": fig["parameterGroupId"],
                "datasetVariableId": subPlot["datasetVariableId"],
                "parameterId": subPlot["parameterId"],
                "label": subPlot["label"],
                "deviceId": subPlot["deviceId"],
            }
            for fig in figs
            for plot in fig["plots"]
            for subPlot in plot["subPlots"]
        }
        metadata["variables_details"] = out
        metadata["variables"] = list(out.keys())

        # include datumConversion info if present
        if len(results["data"]["datumConversions"]) > 0:
            metadata["datumConversions"] = results["data"]["datumConversions"]

        filter = f"%7B%22stations%22:%5B%22{metadata['internal_id']}%22%5D%7D"
        baseurl = "https://sensors.axds.co/api"
        metadata_url = f"{baseurl}/metadata/filter/custom?filter={filter}"
        metadata["metadata_url"] = metadata_url

        # also save units here

        # 1 or 2?
        metadata["version"] = results["data"]["version"]

    return metadata


def make_label(label: str, units: Optional[str] = None, use_units: bool = True) -> str:
    """making column name

    Parameters
    ----------
    label : str
        variable label to use in column header
    units : Optional[str], optional
        units to use in column name, if not None, by default None
    use_units : bool, optional
        Users can choose not to include units in column name, by default True

    Returns
    -------
    str
        string to use as column name
    """

    if units is None or not use_units:
        return f"{label}"
    else:
        return f"{label} [{units}]"


def make_filter(internal_id: int, parameterGroupId: Optional[int] = None) -> str:
    """Make filter for Axiom Sensors API.

    Parameters
    ----------
    internal_id : int
        internal id for station. Not the dataset_id or uuid.
    parameterGroupId : Optional[int], optional
        Parameter Group ID to narrow search, by default None

    Returns
    -------
    str
        filter to use in station metadata and data access
    """

    filter = f"%7B%22stations%22%3A%5B%22{internal_id}%22%5D"

    if parameterGroupId is not None:
        filter += f"%2C%22parameterGroups%22%3A%5B{parameterGroupId}%5D"

    # add ending }
    filter += "%7D"

    return filter


def make_data_url(
    filter: str,
    start_time: str,
    end_time: str,
    binned: bool = False,
    bin_interval: Optional[str] = None,
) -> str:
    """Create url for accessing sensor data, raw or binned.

    Parameters
    ----------
    filter : str
        get this from ``make_filter()``; contains station and potentially variable info.
    start_time : str
        e.g. "2022-1-1". Needs to be interpretable by pandas ``Timestamp``.
    end_time : str
        e.g. "2022-1-2". Needs to be interpretable by pandas ``Timestamp``.
    binned : bool, optional
        True for binned data, False for raw, by default False.
    bin_interval : Optional[str], optional
        If ``binned=True``, input the binning interval to return. Options are hourly, daily, weekly, monthly, yearly.

    Returns
    -------
    str
        URL from which to access data.
    """

    # handle start and end dates (maybe this should happen in cat?)
    start_date = pd.Timestamp(start_time).strftime("%Y-%m-%dT%H:%M:%S")
    end_date = pd.Timestamp(end_time).strftime("%Y-%m-%dT%H:%M:%S")

    if binned:
        return f"{baseurl}/observations/filter/custom/binned?filter={filter}&start={start_date}Z&end={end_date}Z&binInterval={bin_interval}"
    else:
        return f"{baseurl}/observations/filter/custom?filter={filter}&start={start_date}Z&end={end_date}Z"


def make_metadata_url(filter: str) -> str:
    """Make url for finding metadata

    Parameters
    ----------
    filter : str
        filter for Sensors API. Use ``make_filter`` to make this.

    Returns
    -------
    str
        url for metadata.
    """
    return f"{baseurl}/metadata/filter/custom?filter={filter}"


def make_search_docs_url(dataset_id: str) -> str:
    """Url for Axiom Search docs.

    Parameters
    ----------
    dataset_id : str
        dataset_id or uuid.

    Returns
    -------
    str
        Url for finding Axiom Search docs
    """
    return f"https://search.axds.co/v2/docs?verbose=false&id={dataset_id}"


def response_from_url(url: str) -> Union[list, dict]:
    """Return response from url.

    Parameters
    ----------
    url : str
        URL to check.

    Returns
    -------
    list, dict
        should be a list or dict depending on the url
    """
    return requests.get(url, headers=search_headers).json()



"""
Set up a catalog for Axiom assets.
"""


from datetime import datetime
from typing import List, MutableMapping, Optional, Sequence, Tuple, Union

import pandas as pd

from cf_pandas import astype
from intake.catalog.base import Catalog
from intake.catalog.local import LocalCatalogEntry
from intake.source.csv import CSVSource
from intake_parquet.source import ParquetSource

from . import __version__
from .axds import AXDSSensorSource
from .utils import (
    load_metadata,
    match_key_to_parameter,
    match_std_names_to_parameter,
    response_from_url,
)


class AXDSCatalog(Catalog):
    """
    Makes data sources out of all datasets for a given AXDS data type.

    Attributes
    ----------
    pglabels : list[str]
        If ``keys_to_match`` or ``standard_names`` is input to search on, they are converted to parameterGroupLabels and saved to the catalog metadata.
    pgids : list[int]
        If ``keys_to_match`` or ``standard_names`` is input to search on, they are converted to parameterGroupIds and saved to the catalog metadata. In the case that ``query_type=="intersection_constrained"`` and ``datatype=="platform2"``, the pgids are passed to the sensor source so that only data from variables corresponding to those pgids are returned.

    Parameters
    ----------
    datatype : str
        Axiom data type. Currently "platform2" or "sensor_station" but eventually also "module". Platforms and sensors are returned as dataframe containers.
    keys_to_match : str, list, optional
        Name of keys to match with system-available variable parameterNames using criteria. To filter search by variables, either input keys_to_match and a vocabulary or input standard_names. Results from multiple values will be combined according to ``query_type``.
    standard_names : str, list, optional
        Standard names to select from Axiom search parameterNames. If more than one is input, the search is for a logical OR of datasets containing the standard_names. To filter search by variables, either input keys_to_match and a vocabulary or input standard_names. Results from multiple values will be combined according to ``query_type``.
    bbox : tuple of 4 floats, optional
        For explicit geographic search queries, pass a tuple of four floats in the `bbox` argument. The bounding box parameters are `(min_lon, min_lat, max_lon, max_lat)`.
    start_time : str, optional
        For explicit search queries for datasets that contain data after `start_time`. Must include end_time if include start_time.
    end_time : str, optional
        For explicit search queries for datasets that contain data before `end_time`. Must include start_time if include end_time.
    search_for : str, list of strings, optional
        For explicit search queries for datasets that any contain of the terms specified in this keyword argument. Results from multiple values will be combined according to ``query_type``.
    kwargs_search : dict, optional
        Keyword arguments to input to search on the server before making the catalog. Options are:

        * to search by bounding box: include all of min_lon, max_lon, min_lat, max_lat: (int, float). Longitudes must be between -180 to +180.
        * to search within a datetime range: include both of min_time, max_time: interpretable datetime string, e.g., "2021-1-1"
        * to search using a textual keyword: include `search_for` as a string or list of strings. Results from multiple values will be combined according to ``query_type``.

    query_type : str, default "union"
        Specifies how the catalog should apply the query parameters. Choices are:

        * ``"union"``: the results will be the union of each resulting dataset. This is equivalent to a logical OR.
        * ``"intersection"``: the set of results will be the intersection of each individual query made to the server. This is equivalent to a logical AND of the results.
        * ``"intersection_constrained"``: the set of results will be the intersection of queries but also only the variables requested (using either ``keys_to_match`` or ``standard_names``) will be returned in the DataFrame, instead of all available variables. This only applies to ``datatype=="sensor_station"``.

    qartod : bool, int, list, optional
        Whether to return QARTOD agg flags when available, which is only for sensor_stations. Can instead input an int or a list of ints representing the _qa_agg flags for which to return data values. More information about QARTOD testing and flags can be found here: https://cdn.ioos.noaa.gov/media/2020/07/QARTOD-Data-Flags-Manual_version1.2final.pdf. Only used by datatype "sensor_station". Is not available if ``binned==True``.

        Examples of ways to use this input are:

        * ``qartod=True``: Return aggregate QARTOD flags as a column for each data variable.
        * ``qartod=False``: Do not return any QARTOD flag columns.
        * ``qartod=1``: nan any data values for which the aggregated QARTOD flags are not equal to 1.
        * ``qartod=[1,3]``: nan any data values for which the aggregated QARTOD flags are not equal to 1 or 3.

        Flags are:

        * 1: Pass
        * 2: Not Evaluated
        * 3: Suspect
        * 4: Fail
        * 9: Missing Data

    use_units : bool, optional
        If True include units in column names. Syntax is "standard_name [units]". If False, no units. Then syntax for column names is "standard_name". This is currently specific to sensor_station only. Only used by datatype "sensor_station".
    binned : bool, optional
        True for binned data, False for raw, by default False. Only used by datatype "sensor_station".
    bin_interval : Optional[str], optional
        If ``binned=True``, input the binning interval to return. Options are hourly, daily, weekly, monthly, yearly. If bin_interval is input, binned is set to True. Only used by datatype "sensor_station".
    page_size : int, optional
        Number of results. Fewer is faster. Note that default is 10. Note that if you want to make sure you get all available datasets, you should input a large number like 50000.
    verbose : bool, optional
        Set to True for helpful information.
    ttl : int, optional
        Time to live for catalog (in seconds). How long before force-reloading catalog. Set to None to not do this.
    name : str, optional
        Name for catalog.
    description : str, optional
        Description for catalog.
    metadata : dict, optional
        Metadata for catalog.
    kwargs:
        Other input arguments are passed to the intake Catalog class. They can includegetenv, getshell, persist_mode, storage_options, and user_parameters, in addition to some that are surfaced directly in this class.

    Notes
    -----

    only datatype sensor_station uses the following parameters: qartod, use_units, binned, bin_interval

    datatype of sensor_station skips webcam data.

    """

    name = "axds_cat"
    version = __version__

    def __init__(
        self,
        datatype: str,
        keys_to_match: Optional[Union[str, list]] = None,
        standard_names: Optional[Union[str, list]] = None,
        bbox: Optional[Tuple[float, float, float, float]] = None,
        start_time: Optional[str] = None,
        end_time: Optional[str] = None,
        search_for: Optional[Union[str, List[str]]] = None,
        kwargs_search: MutableMapping[
            str, Union[str, int, float, Sequence[Union[str, None]]]
        ] = None,
        query_type: str = "union",
        qartod: Union[bool, int, List[int]] = False,
        use_units: bool = True,
        binned: bool = False,
        bin_interval: Optional[str] = None,
        page_size: int = 10,
        verbose: bool = False,
        name: str = "catalog",
        description: str = "Catalog of Axiom assets.",
        metadata: dict = None,
        ttl: Optional[int] = None,
        **kwargs,
    ):

        self.datatype = datatype
        self.kwargs_search = kwargs_search
        self.page_size = page_size
        self.verbose = verbose
        self.qartod = qartod
        self.use_units = use_units
        self.kwargs_search = kwargs_search or {}
        self.query_type = query_type

        if bin_interval is not None:
            binned = True
        self.binned = binned
        self.bin_interval = bin_interval

        if self.binned:
            if self.qartod:
                raise ValueError(
                    "QARTOD is not available for binned output. Set QARTOD to False or use raw data."
                )

        allowed_datatypes = ("platform2", "sensor_station")
        if datatype not in allowed_datatypes:
            raise KeyError(
                f"Datatype must be one of {allowed_datatypes} but is {datatype}"
            )

        allowed_query_types = ("union", "intersection", "intersection_constrained")
        if query_type not in allowed_query_types:
            raise KeyError(
                f"`query_type` must be one of {allowed_query_types} but is {query_type}"
            )
        if (
            query_type == "intersection" or query_type == "intersection_constrained"
        ) and page_size == 10:
            if verbose:
                print(
                    "With `query_type` of 'intersection' you probably want to use a larger `page_size` than the default of 10."
                )

        if query_type == "intersection_constrained" and datatype == "platform2":
            raise ValueError(
                "`query_type=='intersection_constrained'` does not apply to `datatype=='platform2'`."
            )

        # can instead input the kwargs_search outside of that dictionary
        if bbox is not None:
            if not isinstance(bbox, tuple):
                raise TypeError(
                    f"Expecting a tuple of four floats for argument bbox: {type(bbox)}"
                )
            if len(bbox) != 4:
                raise ValueError("bbox argument requires a tuple of four floats")
            self.kwargs_search["min_lon"] = bbox[0]
            self.kwargs_search["min_lat"] = bbox[1]
            self.kwargs_search["max_lon"] = bbox[2]
            self.kwargs_search["max_lat"] = bbox[3]

        if start_time is not None:
            if start_time in self.kwargs_search:
                raise KeyError("start_time defined explicitly and in kwargs_search.")

            if not isinstance(start_time, (str, datetime)):
                raise TypeError(
                    f"Expecting a datetime for start_time argument: {repr(start_time)}"
                )
            # if isinstance(start_time, str):
            #     start_time = pd.Timestamp(start_time)#.strptime(start_time, "%Y-%m-%dT%H:%M:%SZ")
            self.kwargs_search["min_time"] = start_time

        if end_time is not None:
            if end_time in self.kwargs_search:
                raise KeyError("end_time defined explicitly and in kwargs_search.")
            if not isinstance(end_time, (str, datetime)):
                raise TypeError(
                    f"Expecting a datetime for end_time argument: {repr(end_time)}"
                )
            # if isinstance(end_time, str):
            #     end_time = pd.Timestamp(end_time)#.strptime(end_time, "%Y-%m-%dT%H:%M:%SZ")
            self.kwargs_search["max_time"] = end_time

        if search_for is not None:
            if "search_for" in self.kwargs_search:
                raise KeyError("search_for defined explicitly and in kwargs_search.")
            if not isinstance(search_for, (str, list)):
                raise TypeError(
                    f"Expecting string or list of strings for search_for argument: {repr(search_for)}"
                )
            self.kwargs_search["search_for"] = search_for
        if "search_for" not in self.kwargs_search:
            self.kwargs_search["search_for"] = [None]
        else:
            if isinstance(self.kwargs_search["search_for"], str):
                self.kwargs_search["search_for"] = [self.kwargs_search["search_for"]]

        checks = [
            ["min_lon", "max_lon", "min_lat", "max_lat"],
            ["min_time", "max_time"],
        ]
        for check in checks:
            if any(key in self.kwargs_search for key in check) and not all(
                key in self.kwargs_search for key in check
            ):
                raise ValueError(
                    f"If any of {check} are input, they all must be input."
                )

        if "min_lon" in self.kwargs_search and "max_lon" in self.kwargs_search:
            min_lon, max_lon = (
                self.kwargs_search["min_lon"],
                self.kwargs_search["max_lon"],
            )
            if isinstance(min_lon, (int, float)) and isinstance(max_lon, (int, float)):
                if abs(min_lon) > 180 or abs(max_lon) > 180:
                    raise ValueError(
                        "`min_lon` and `max_lon` must be in the range -180 to 180."
                    )

        # input keys_to_match OR standard_names but not both
        if keys_to_match is not None and standard_names is not None:
            raise ValueError(
                "Input either `keys_to_match` or `standard_names` but not both."
            )

        self.pglabels: List[Union[str, None]]
        self.pgids: List[Union[int, None]]
        if keys_to_match is not None:
            self.pglabels, self.pgids = match_key_to_parameter(
                astype(keys_to_match, list)
            )
        elif standard_names is not None:
            self.pglabels, self.pgids = match_std_names_to_parameter(
                astype(standard_names, list)
            )
        else:
            self.pglabels, self.pgids = [None], [None]

        # Put together catalog-level stuff
        if metadata is None:
            metadata = {}
            metadata["kwargs_search"] = self.kwargs_search
            metadata["pglabels"] = list(self.pglabels)
            metadata["pgids"] = list(self.pgids)
            metadata["query_type"] = query_type
            # metadata["use_units"] = use_units

        super(AXDSCatalog, self).__init__(
            **kwargs, ttl=ttl, name=name, description=description, metadata=metadata
        )

    def search_url(
        self, pglabel: Optional[str] = None, text_search: Optional[str] = None
    ) -> str:
        """Set up one url for searching.

        Parameters
        ----------
        pglabel : Optional[str], optional
            Parameter Group Label (not ID), by default None
        text_search : Optional[str], optional
            free text search, by default None

        Returns
        -------
        str
            URL to use to search Axiom systems.
        """

        self.url_search_base = f"https://search.axds.co/v2/search?portalId=-1&page=1&pageSize={self.page_size}&verbose=true"

        url = f"{self.url_search_base}&type={self.datatype}"

        assert isinstance(self.kwargs_search, dict)
        if self.kwargs_search.keys() >= {
            "max_lon",
            "min_lon",
            "min_lat",
            "max_lat",
        }:
            url_add_box = (
                f'&geom={{"type":"Polygon","coordinates":[[[{self.kwargs_search["min_lon"]},{self.kwargs_search["min_lat"]}],'
                + f'[{self.kwargs_search["max_lon"]},{self.kwargs_search["min_lat"]}],'
                + f'[{self.kwargs_search["max_lon"]},{self.kwargs_search["max_lat"]}],'
                + f'[{self.kwargs_search["min_lon"]},{self.kwargs_search["max_lat"]}],'
                + f'[{self.kwargs_search["min_lon"]},{self.kwargs_search["min_lat"]}]]]}}'
            )
            url += f"{url_add_box}"

        if self.kwargs_search.keys() >= {"max_time", "min_time"}:
            # convert input datetime to seconds since 1970
            startDateTime = (
                pd.Timestamp(self.kwargs_search["min_time"]).tz_localize("UTC")
                - pd.Timestamp("1970-01-01 00:00").tz_localize("UTC")
            ) // pd.Timedelta("1s")
            endDateTime = (
                pd.Timestamp(self.kwargs_search["max_time"]).tz_localize("UTC")
                - pd.Timestamp("1970-01-01 00:00").tz_localize("UTC")
            ) // pd.Timedelta("1s")

            # search by time
            url_add_time = f"&startDateTime={startDateTime}&endDateTime={endDateTime}"

            url += f"{url_add_time}"

        if text_search is not None:
            url += f"&query={text_search}"

        # search by variable
        if pglabel is not None:
            url += f"&tag=Parameter+Group:{pglabel}"

        # if requests.get(url).status_code != 200:
        #     raise ValueError("")

        if self.verbose:
            print(f"search url: {url}")

        return url

    def get_search_urls(self) -> list:
        """Gather all search urls for catalog.

        Inputs that can have more than one search_url are pglabels and search_for list.

        Returns
        -------
        list
            List of search urls.
        """
        assert isinstance(self.kwargs_search, dict)
        search_urls = [
            self.search_url(pglabel, text_search)
            for pglabel in self.pglabels
            for text_search in self.kwargs_search["search_for"]
        ]

        return search_urls

    def _load_all_results(self) -> list:
        """Combine results from multiple search urls using query_type.

        Returns
        -------
        list
            list of results from potentially multiple searches.

        Raises
        ------
        ValueError
            if no results found.
        """

        combined_results = []
        first_loop = True
        for search_url in self.get_search_urls():
            res = response_from_url(search_url)
            if "results" not in res:
                raise ValueError(
                    f"No results were returned for the search. Search url: {search_url}."
                )

            assert isinstance(res, dict)

            if self.verbose:
                print(
                    f"For search url {search_url}, number of results found: {len(res['results'])}. Page size: {self.page_size}."
                )

            if self.query_type == "union":  # logical OR
                combined_results.extend(res["results"])

            elif (
                self.query_type == "intersection"
                or self.query_type == "intersection_constrained"
            ):  # logical AND
                if first_loop:  # initialize
                    combined_results = res["results"]
                    first_loop = False
                else:
                    # compare uuids from first search with uuids from next search
                    uuids_before = [dataset["uuid"] for dataset in combined_results]
                    uuids_now = [dataset["uuid"] for dataset in res["results"]]

                    overlapping_uuids = set(uuids_before).intersection(set(uuids_now))

                    combined_results = [
                        dataset
                        for dataset in combined_results
                        if dataset["uuid"] in overlapping_uuids
                    ]

        if self.verbose:
            unique = set([res["uuid"] for res in combined_results])
            print(
                f"Total number of results found for page_size {self.page_size} over {len(self.get_search_urls())} different searches with query_type {self.query_type}: {len(combined_results)}, with unique results: {len(unique)}."
            )

        return combined_results

    def _load(self):
        """Find all dataset ids and create catalog."""

        results = self._load_all_results()

        self._entries = {}
        for result in results:
            dataset_id = result["uuid"]

            # don't repeat an entry (it won't actually allow you to, but probably saves time not to try)
            if dataset_id in self._entries:
                continue

            if self.verbose:
                print(f"Dataset ID: {dataset_id}")

            # # quick check if OPENDAP is in the access methods for this uuid, otherwise move on
            # if self.datatype == "module":
            #     # if opendap is not in the access methods at the module level, then we assume it
            #     # also isn't at the layer_group level, so we will not check each layer_group
            #     if "OPENDAP" not in results["data"]["access_methods"]:
            #         if self.verbose:
            #             print(
            #                 f"Cannot access module {dataset_id} via opendap so no source is being made for it.",
            #                 UserWarning,
            #             )
            #         # warnings.warn(f"Cannot access module {dataset_id} via opendap so no source is being made for it.", UserWarning)
            #         continue
            #     if "DEPRECATED" in results["data"]["label"]:
            #         if self.verbose:
            #             print(
            #                 f"Skipping module {dataset_id} because label says it is deprecated.",
            #                 UserWarning,
            #             )
            #         continue

            description = f"AXDS dataset_id {dataset_id} of datatype {self.datatype}"

            metadata = load_metadata(self.datatype, result)

            # don't save Camera sensor data for now
            if "webcam" in metadata["variables"]:
                if self.verbose:
                    print(f"Dataset_id {dataset_id} is a webcam so is being skipped.")
                continue

            # Find urlpath
            if self.datatype == "platform2":
                # use parquet if available, otherwise csv
                try:
                    key = [
                        key
                        for key in result["source"]["files"].keys()
                        if ".parquet" in key
                    ][0]
                    urlpath = result["source"]["files"][key]["url"]
                    plugin = ParquetSource
                except Exception:
                    urlpath = result["source"]["files"]["data.csv.gz"]["url"]
                    plugin = CSVSource

                args = {
                    "urlpath": urlpath,
                }

            # this Source has different arg requirements
            elif self.datatype == "sensor_station":
                args = {
                    "dataset_id": dataset_id,
                    "internal_id": metadata["internal_id"],
                    "start_time": self.kwargs_search.get("min_time", None),
                    "end_time": self.kwargs_search.get("max_time", None),
                    # "kwargs_search": self.kwargs_search,
                    "qartod": self.qartod,
                    "use_units": self.use_units,
                    "binned": self.binned,
                    "bin_interval": self.bin_interval,
                    "only_pgids": list(self.pgids)
                    if self.query_type == "intersection_constrained"
                    else None,
                }
                plugin = AXDSSensorSource

            # elif self.datatype == "module":
            #     plugin = NetCDFSource  # 'netcdf'

            #     # modules are the umbrella and contain 1 or more layer_groups
            #     # pull out associated layer groups uuids to make sure to capture them
            #     layer_group_uuids = list(docs["data"]["layer_group_info"].keys())

            #     # pull up docs for each layer_group to get urlpath
            #     # can only get a urlpath if it is available on opendap
            #     urlpaths = []  # using this to see if there are ever multiple urlpaths
            #     for layer_group_uuid in layer_group_uuids:
            #         docs_lg = return_docs_response(layer_group_uuid)
            # docs_log = response_from_url(make_search_docs_url(layer_group_uuid))

            #         if "OPENDAP" in docs_lg["data"]["access_methods"]:
            #             urlpath = docs_lg["source"]["layers"][0][
            #                 "thredds_opendap_url"
            #             ].removesuffix(".html")
            #             urlpaths.append(urlpath)

            #     # only want unique urlpaths
            #     urlpaths = list(set(urlpaths))
            #     if len(urlpaths) > 1:
            #         if self.verbose:
            #             print(
            #                 f"Several urlpaths were found for module {dataset_id} so no source is being made for it."
            #             )
            #         # warnings.warn(f"Several urlpaths were found for module {dataset_id} so no source is being made for it.", UserWarning)
            #         continue
            #         # raise ValueError(f"the layer_groups for module {dataset_id} have different urlpaths.")
            #     elif len(urlpaths) == 0:
            #         if self.verbose:
            #             print(
            #                 f"No urlpath was found for module {dataset_id} so no source is being made for it."
            #             )
            #         # warnings.warn(f"No urlpath was found for module {dataset_id} so no source is being made for it.", UserWarning)
            #         continue
            #     else:
            #         urlpath = urlpaths[0]

            entry = LocalCatalogEntry(
                name=dataset_id,
                description=description,
                driver=plugin,
                direct_access="allow",
                args=args,
                metadata=metadata,
                # True,
                # args,
                # {},
                # {},
                # {},
                # "",
                # getenv=False,
                # getshell=False,
            )

            entry._plugin = [plugin]

            self._entries[dataset_id] = entry
